/*
  purpose: converts a pnm file to C source code for Nokia 5110/6110 displays

  author: Henryk Richter <buggs@comlabien.net>

  compilation: gcc -o pnm2nokia pnm2nokia.c

  syntax: pnm2nokia <infile> <outfile>

  inputs: binary PNM/PPM files (types P4,P5,P6, 8 Bit only for the latter two)

  output: 
   - C-formatted file, dimensions of image as comment 

  notes:
   - the 5510 display has a display size of 84*48 pixels
   - each byte in the display buffer is 1*8 pixels (1 column, 8 rows)
   - hence there are 6 rows of 8 pixels each within each column
   - consequences:
     - images where the vertical size is not a multiple of 8 are
       zero padded 
     - images larger than 84x48 are cropped
     - horizontal size is flexible 1...84 pixels
   
   data organization:
   - column by column, starting with the upleft corner,
     where each byte consists of 8 vertically adjacent pixels
   - image height is still given in pixel units, your program must
     compute the number of display rows by itself (or use buffered
     display output where the exact number of pixel rows can be implemented
     by some bit-shifting)

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BAILOUT(a) fclose(in);fclose(out);printf(a);exit(0)

/*---------------- convert input pixmap (8 Bit Gray) to Nokia 5110 -------------*/
#define N5110_MAXW 84
#define N5110_MAXH 48
int img2nokia( unsigned char *buf, long w, long h, FILE *out )
{
	int i,j,mw,mh,r8,k;
	unsigned int o,t,c;

	/* crop ? */
	mw = (w > N5110_MAXW) ? N5110_MAXW : w;
	mh = (h > N5110_MAXH) ? N5110_MAXH : h;
	r8 = (mh+7)/8; /* round up to n*8 rows, in 8 pixel units */

	fprintf(out,"/* File generated by pnm2nokia                         */\n");
	fprintf(out,"/* http://bax.comlab.uni-rostock.de                    */\n");
	fprintf(out,"/*                                                     */\n\n");
	fprintf(out,"/* nokia image width %d height %d */\n",mw,mh);
	fprintf( out, "const unsigned char n5110_image[%d] = {\n",mw*r8);

	for( j = 0, c=0 ; j < r8 ; j++ ) /* round up to n*8 rows, in 8 pixel units */
	{
		for( i=0 ; i < mw ; i++ )
		{
			/* get 8 vertically adjacent pixels */
			o=0;
			for( k=0 ; k<8 ; k++ )
			{
				t = (( 128 & buf[ (j*8+k)*w +i ] ) >> 7)^1; /* thresholding: >128 = 0, else 1 */
				o |= (t << k); /* first row is LSB, last row MSB */
			}
			
			fprintf( out, "0x%02x,",o );

			c++;
			if( (c & 15) == 0 )
			fprintf( out,"\n");

		}
	}
	fprintf( out, "};\n");

	return 0;
}



/*---------------- read one ASCII line and reject comment lines ----------------*/
int readline(FILE *infile, char * string)
{
 int aha,i;
 char t,*str;

 do
 {
  str = string;
  aha=1;
  i = 0;

  do
   {
    t = fgetc(infile);
    *str++ = t;

    if( t == '#' )
     aha=0;

    i++;
   }
  while( (i<200) && (t != '\n') && (t != EOF) );

  *(str-1) = 0;
 }
 while( (aha == 0) && (i<200) );

 if(i >= 200)
  aha = 0;

 return(aha);
}

/*------------------------main loop-----------------------------------------------*/
int main( int argc, char* argv[] )
{
 FILE *in=(0),*out=(0);
 char stringbuf[200];
 long w,h,t;
 int mode,tmp;

	if( argc < 3 )
	{
		printf("usage: %s infile.pnm outfile.c\n",argv[0] );
		exit(0);
	}

 in =  fopen( argv[1], "rb" );
 if( in == NULL )
 {
	printf("cannot open input file %s\n", argv[1] );
	exit(0);
 }

 out = fopen( argv[2], "wb" );
 if( out == NULL )
 {
	printf("cannot open input file %s\n", argv[2] );
	fclose( in );
	exit(0);
 }

 mode = 0;
 readline(in,stringbuf);
 if( 0 != strncmp( stringbuf, "P6",2 ) )
 {
	if( 0 == strncmp( stringbuf, "P5",2 ) )
	{
			mode = 1;
	}
	else
	{
	 	if( 0 != strncmp( stringbuf, "P4",2 ) )
		{
			BAILOUT("Sorry, not 1/8/24 bit binary pnm (P4/P5/P6)\n");
		}
		mode = 2;
 	}
 }

 /* check if all arguments in one line */
 if( 4 != sscanf( stringbuf, "P%1d %ld %ld %ld",&tmp,&w,&h,&t ) )
 {
  readline(in,stringbuf);
  if( 1 == sscanf(stringbuf,"%ld %ld",&w,&h)) /* 1 or 2 words in line ? */
  {
    if( readline(in,stringbuf) )     /* if 1 in last line, fetch next line */
     sscanf(stringbuf,"%ld",&h);
  }

  readline(in,stringbuf);	     /* result: 255 */
 }

 printf("w is %ld, h is %ld mode is %d\n",w,h,mode);
 if( (mode == 2) && (w&7) )
 {
	printf("Warning: width in PBM is not a multiple of 8, expect problems\n");
 }

// fprintf( out, "unsigned int tex_w=%ld;\n",w);
// fprintf( out, "unsigned int tex_h=%ld;\n\n",w);
// fprintf( out, "unsigned char tex_data[] = {\n");

 {
  unsigned char a;
  int  i,max,j;
  unsigned char *buf,*bptr;
  unsigned char b1[3]; /* RGB buffer */
  
  max = w*h;
  buf = (unsigned char*)malloc( w*(h+7) ) ; /* 1 byte per pixel, sufficient space for zero-padding */
  if( !buf )
  {
  	BAILOUT("Cannot allocate memory buffer\n");
  }
  memset( buf, 0, w*(h+7) );

  /* read file, convert to 8 bit gray */
  switch(mode)
  {
	case 0:
		/* RGB */
		for( i = 0 ; i < max ; i++ )
	  	{
			fread( b1, 1, 3, in );
			buf[i] = ( (int)b1[0] + (int)b1[1] + (int)b1[2]) / 3; /* RGB2GRAY */
		}
		break;
	case 1: 
		/* Gray */
		fread( buf, 1, max, in );
		break;
	case 2: 
		/* Bitmap (converted to 8 bit gray) */
	 	max = (max+7)/8; /* round up to n*8 Bits */
		bptr= buf;
		for( i = 0 ; i < max ; i++ )
	  	{
			fread( &a, 1, 1, in );
			for( j = 128; j != 0 ; j >>= 1 )
			{
				*bptr++ = (a&j) ? 255 : 0;
			}
		}
		break;
	default:
		printf("Error: wrong internal mode, this shouldn't have happened...\n");
		break;
  }

  /* now write the output file */
  img2nokia( buf, w, h, out );

  free(buf);
 }
#if 0
  if( !mode )   max *= 3;
  if( mode==2 ) 
  {
  	int j;

 	  max = (max+7)/8; /* round up to n*8 Bits */

	  for( i = 0 ; i < max ; i++ )
	  {
	   fread( &a, 1, 1, in );
	   for( j = 128; j != 0 ; j >>= 1 )
	   {
	   	fprintf( out, "0x%02x,", (a&j) ? 255 : 0 );
	   }

	   if( ((i % 2) == 0) && (i>0) )
		fprintf( out,"\n");
	  }
 
  }
  else
  {
	  for( i = 0 ; i < max ; i++ )
	  {
	   fread( &a, 1, 1, in );
	   fprintf( out, "0x%02x,",a );

	   if( ((i % 2) == 0) && (i>0) )
		fprintf( out,"\n");
	  }
  }
 }

 fprintf( out, "};\n");
#endif

error:
 fclose(in);
 fclose(out);

 return 0;
}
