/*
 ********************************************************************************
 * main.cpp                                                                     *
 *                                                                              *
 * Author: Henryk Richter <bax@comlab.uni-rostock.de>                           *
 *                                                                              *
 * Purpose: Nokia 6110/5110 display control plus UART comms                     *
 *                                                                              *
 ********************************************************************************
*/
/* Pins on Atmel / LCD
Layout assumed here (for my Atmega328p)
- PB0 D/C 
- PB1 SS (SCE)
  PB2 SS (of atmel, unused)
- PB3 MOSI (DATA)
- PB4 MISO (-> RST, important: careful with the Pin settings when issuing reset, may break HW SPI when used )
- PB5 SCK (CLK)
*/


#include <avr/interrupt.h>
//#include <avr/io.h>
#include "timeout.h"
#include "5110.h"
#include "uart.h"

/* graphics */
#include "img_prefixes.h"
#include "img_hearts.h"
#include "img_title.h"

#if 1
#define DBGOUT(a) uart_putc(a);
#else
#define DBGOUT(a)
#endif

/* C++ 5110 LCD library (modified) */
LCD_5110 lcd;


/* config of displayed items (number 1...12) 
   single character per display (program iterates to proper position)
   - small chars: one of two columns occupied
   - caps chars:  full line 
   - t/T = text field without graphic prefix
   - f/F = fan image
   - g/G = GPU image
   - c/C = CPU image
   - u/U = Clock image
   - m/M = memory chip
   - z/Z = clock cycles

   - Serial Protocol:
     - 'A'-'Z': new args for display 0..n, \r delimits single argument (string, e.g. "a12.4V\r")
     - '@': config, delimited by \r
     - \r : ignored, resync
     - a-z: echoed (ignored otherwise, ignore mode ends with \r)
*/
char animidx=0; // animated prefix icons
#define DISP_MAXITEMS 12
#define DISP_HSTRIDE  42
char disp_nitems = 0; // set to 7 for testing
char disp_config[DISP_MAXITEMS] = {'T','f','f','c','g','U','B',0,0,0,0,0};

void display_configure( void );
void draw_prefixes( char animidx );

int main(void)
{
  int           idlecount; /* check whether we are idle (no PC client active) */
  unsigned char heartbeat; /* heartbeat counter                               */

  _delay_ms(100);  /* wait for display to settle in */

  /* LCD Data Pins (configurable): SCE, RST, DC, DATA, CLK */
  lcd.lcd_init(&PORTB, PB1, &PORTB, PB4, &PORTB, PB0, &PORTB, PB3, &PORTB, PB5);

  /* init UART for serial communication to PC */
  uart_init(UART_BAUD_SELECT(9600UL,16000000UL)); /* FIXME: F_CPU */
  sei();
  uart_puts("LCD SYSTEM STATUS 5110 BY BAX\r\n");

  /* start with a default picture */
  lcd.drawPictureAt( img_title, 0, 0, 84, 48 );

  /* main loop */
  idlecount = -1;
  heartbeat = 0;
  while(1)
  {
  		char a;

		/* begin idle/heartbeat considerations */
		if( idlecount >= 0 )
			idlecount++; /* yep, the sign bit _will_ flip after a while, this is intended,see below */
		if( heartbeat == 64 )
		{
			lcd.drawPictureAt( (idlecount<0) ? n5110_image_h1 : n5110_image_h3, 78, 0, 5, 8 );
		}
		if( heartbeat >= 128 )
		{
			lcd.drawPictureAt( (idlecount<0) ? n5110_image_h2 : n5110_image_h4, 78, 0, 5, 8 );
			heartbeat = 0;
			if( idlecount < 0 )
				uart_puts("LCD SYSTEM STATUS 5110 BY BAX\r\n"); /* notify host that we are alive */
		}
		if( !(heartbeat & 31) ) /* animate prefix graphics */
		{
			animidx += IMAGE_PREFIXES_ANIMSTEP; 
			if( animidx > IMAGE_PREFIXES_ANIMMAX ) 
				animidx = 0;
			draw_prefixes( animidx );
		}
		/* end idle/heartbeat considerations */

		/* new (complete) command ? */
		if( uart_have_newline() > 0 )
		{
			idlecount = 0;

			a = uart_getc();
			DBGOUT(a)
			if( a == '@' )
			{
				uart_puts("config mode ");
				/* configure request */
				display_configure();
				uart_puts("\r\n");
			}
			else /* if( a == '@' ) */
			{
				if( ( a >= 'A' ) && (a < ('A'+disp_nitems) ))
				{
					unsigned char i=0,x=0,y=0;
					a-='A';
					/* display request:calculate position on screen */
					for( i=0; i < a; i++ )
					{
						if( (disp_config[i] < 'a') || (x>0) ) /* just 2 columns supported, next line if second column */
						{
							x  = 0;
							y += 1;
						}
						else
							x += DISP_HSTRIDE;
					}
					if( disp_config[i] != 'T' ) /* full text line ? */
						x += IMAGE_PREFIXES_ANIMSTEP+1; /* no, move right */

					/* display string at calculated position */
					lcd.lcd_goto_xy_exact( x, y );
					if( disp_config[i] == 'b' )
					{
						lcd.lcd_base_addr( (int)x + (int)84*(int)y );
						x=0;
						/* display bar */
						a=uart_getc();
						while( a > 13 )
						{
							x *= 10;
							x += a-'0';
							a=uart_getc(); /* next char */
						}
						x >>= 2; /* 0...100 -> 0..25 */

						lcd.lcd_single_col( 0x7e );
						for( i=0 ; i < x ; i++ )
							lcd.lcd_single_col( 0x7e );
						for(     ; i < 25; i++ )
							lcd.lcd_single_col( 0x42 );
						lcd.lcd_single_col( 0x7e );
					}
					else
					{
						lcd.lcd_str((char*)"     ");
						lcd.lcd_goto_xy_exact( x, y );
						i=0;
						do
						{
							i++;
							a=uart_getc(); /* \n is ignored in display routine */
							DBGOUT(a)
							lcd.lcd_chr( a );
						}
						while( a > 13 );
					}
				}
				else
				{
					uart_puts("ignore mode ");
					/* ignored request (echoed) */
					while( a > 13 )
					{
						uart_putc( a );
						a=uart_getc();
					}
					uart_putc( 13 );
					uart_putc( 10 );
				}
			} /* if( a == '@' ) */
		}	/* new (complete) command ? */
#if 0
		/* UART Loopback test */
		while( uart_available() > 0 )
		{
			char a;

			a = uart_getc();
			uart_putc( a );
			idlecount=0;
		}
#endif
	heartbeat++;
	if( idlecount != 0 )
		_delay_ms(10); /* wait 10ms if nothing important to do */
  }		
}

/* draw prefixes */
void draw_prefixes( char animidx )
{
	unsigned char a,i,x=0,y=0;
	/* display request:calculate position on screen */
	for( i=0; i < disp_nitems; i++ )
	{
		a = disp_config[i] | 32;
		switch( a )
		{
			case 'f':	a=0;	break; /* fan */
			case 'c':	a=1;	break; /* cpu */
			case 'm':	a=2;	break; /* mem */
			case 'u':	a=3;	break; /* wall clock */
			case 'z':	a=4;	break; /* clock cycles */
			case 'g':	a=5;	break; /* gpu */
			default:		break;
		}
		if( a < IMAGE_PREFIXES_COUNT )
			lcd.drawPictureAt( img_prefixes + (((int)(a))<<IMAGE_PREFIXES_STRIDEBITS) + animidx, x,y, IMAGE_PREFIXES_ANIMSTEP,8);

		if( (disp_config[i] < 'a') || (x>0) ) /* just 2 columns supported, next line if second column */
		{
			x  = 0;
			y += 8;
		}
		else
			x += DISP_HSTRIDE;
	}

}

void display_configure( void )
{
	char a;

	disp_nitems = 0;
	a=uart_getc(); /* skip header */
	DBGOUT(a)
	while( (a > 13) && (a!=0) ) /* safety: stop if no more data from UART */
	{
		if( disp_nitems < DISP_MAXITEMS )
		{
			*(disp_config + disp_nitems) = a;
			disp_nitems++;
		}
		a=uart_getc();
		DBGOUT(a)
	}
	lcd.lcd_clear();
	draw_prefixes(0);
}

